open Match_unif
open Utils


type var = int;;
type symb = string;;
type def_symbs = symb list;;

type rule = term * term;;
type system = rule list;;
type dp = term * term;;

(* On utilise les graphes d'ocamlgraph. Comme ceux-ci ne permettent pas
   d'avoir accès au sommet associé à chaque élément, on ajoutant des tables
   de hachage compensant ce problème. *)

type graph = {graph : Graph.Pack.Digraph.t; vertex_of_dp : (dp, Graph.Pack.Digraph.V.t) Hashtbl.t; dp_of_vertex : (Graph.Pack.Digraph.V.t, dp) Hashtbl.t};;

type can_be_removed = Strict | Large | No;;


(* Entier indiquant la pronfondeur de réduction maximum pour savoir si
   on peut retirer une paire de dépendance d'un graphe. *)

let max_test = 20;;


(* Calcul de l'ensemble des symboles définis d'un système de
   récriture *)

let def_symbs_of_rule r liste =
  let (l,_) = r in
  match l with
    | Var _ -> liste
    | Term(f,_) when not (List.mem f liste) -> f::liste
    | Term(_,_) -> liste;;


let rec compute_def_symbs (s:system) : def_symbs =
  match s with
    | [] -> []
    | r::q -> (def_symbs_of_rule r (compute_def_symbs q));;



(* Calcul de l'ensemble des symboles définis avec arité d'un système de
   récriture *)

let def_symbs_arit_of_rule r liste =
  let (l,_) = r in
  match l with
    | Var _ -> liste
    | Term(f,ll) ->
        let n = List.length ll in
        if (List.mem (f,n) liste) then
          liste
        else
          (f,n)::liste;;


let rec compute_def_symbs_arit (s:system) : (symb*int) list =
  match s with
    | [] -> []
    | r::q -> (def_symbs_arit_of_rule r (compute_def_symbs_arit q));;


(* Calcul de l'ensemble des symboles d'un système de récriture *)

let rec symbs_of_term t =
  match t with
    | Var _ -> []
    | Term(f,tl) -> f::(List.flatten (List.rev_map symbs_of_term tl));;


let symbs_of_rule r =
  let (g,d) = r in
  List.rev_append (symbs_of_term g) (symbs_of_term d);;


let compute_symbs (s:system) : symb list =

  let rec rec_compute_symbs s acc =
    match s with
      | [] -> acc
      | r::q -> rec_compute_symbs q (List.rev_append (symbs_of_rule r) acc) in

  superfluous_list (rec_compute_symbs s []);;


(* Calcul de l'ensemble des symboles d'un système de récriture avec arité *)

let rec symbs_arit_of_term t =
  match t with
    | Var _ -> []
    | Term(f,tl) -> (f,List.length tl)::(List.flatten (List.rev_map symbs_arit_of_term tl));;


let symbs_arit_of_rule r =
  let (g,d) = r in
  List.rev_append (symbs_arit_of_term g) (symbs_arit_of_term d);;


let compute_symbs_arit (s:system) : (symb*int) list =

  let rec rec_compute_symbs_arit s acc =
    match s with
      | [] -> acc
      | r::q -> rec_compute_symbs_arit q (List.rev_append (symbs_arit_of_rule r) acc) in

  superfluous_list (rec_compute_symbs_arit s []);;


(* Calcul de l'ensemble des symboles constructeurs d'un système de
   récriture *)

let compute_const_symbs (s:system) : symb list =
  let def_symbs = compute_def_symbs s
  and symbs = compute_symbs s in

  List.filter (fun x -> not (List.mem x def_symbs)) symbs;;


(* Calcul de l'ensemble des paires de dépendance d'un système de
   récriture *)

let rec dps_of_term g ds d =
  match d with
    | Var _ -> []
    | Term(f,l) when (List.mem f ds) -> (g,d)::(List.flatten (List.rev_map (dps_of_term g ds) l))
    | Term(_,l) -> List.flatten (List.rev_map (dps_of_term g ds) l)


let dps_of_rule r ds =
  let (g,d) = r in
  dps_of_term g ds d;;


let compute_dps (s:system) : dp list =
  let ds = compute_def_symbs s in

  let rec rec_compute_dps s acc =
    match s with
      | [] -> acc
      | r::q -> rec_compute_dps q (List.rev_append (dps_of_rule r ds) acc) in

  rec_compute_dps s [];;


(* Calcul du maximum des variables d'un terme *)

let rec max_term t =
  match t with
    | Var v -> v
    | Term(_,l) -> max_list l

and max_list l =
  match l with
    | [] -> -1
    | t::q -> max (max_term t) (max_list q);;


(* Fonctions REN et CAP *)

let cap (const_symbs: symb list) (t:term) : term =

  let cur_var = ref (max_term t) in

  let new_var () =
    incr cur_var;
    !cur_var in

  let rec cap_aux t =
    match t with
      | Var v -> Var v
      | Term(f,l) when List.mem f const_symbs -> Term(f,List.map cap_aux l)
      | Term(f,l) -> Var (new_var ()) in

  match t with
    | Var v -> Var v
    | Term(f,l) -> Term(f,List.map cap_aux l);;


let ren (t:term) (t':term) : term =

  let cur_var = ref (max_term t') in

  let new_var () =
    incr cur_var;
    !cur_var in

  let rec ren_aux t =
    match t with
      | Var v -> Var (new_var ())
      | Term(f,l) -> Term(f,List.map ren_aux l) in

  ren_aux t;;


(* Calcul d'une sur-approximation du graphe de dépendances *)

let compute_graph (const_symbs:symb list) (dp_list:dp list) : graph =
  let n = List.length dp_list in
  let g = Graph.Pack.Digraph.create ~size:n () in
  let v_of_dp = Hashtbl.create n
  and dp_of_v = Hashtbl.create n in

  (* Création des sommets *)
  (* On associe un sommet à chaque paire de dépendance *)

  let count = ref (-1) in
  List.iter (fun dep ->
               incr count;
               let v = Graph.Pack.Digraph.V.create !count in
               Graph.Pack.Digraph.add_vertex g v;
               Hashtbl.add v_of_dp dep v;
               Hashtbl.add dp_of_v v dep) dp_list;

  (* Création des arêtes *)

  List.iter (fun dep1 ->
               let (s,t) = dep1 in
               let v1 = Hashtbl.find v_of_dp dep1 in
               List.iter (fun dep2 ->
                            let (s',t') = dep2 in
                            let v2 = Hashtbl.find v_of_dp dep2 in
                            match unification (ren (cap const_symbs t) s') s' with
                              | Some _ -> Graph.Pack.Digraph.add_edge g v1 v2
                              | None -> ()) dp_list) dp_list;

  {graph = g; vertex_of_dp = v_of_dp; dp_of_vertex = dp_of_v};;


(* Fonction qui calcule le sous-graphe d'un graphe à partir d'une liste
   de sommets *)

let subgraph_of_vertex_list g l =
  let n = List.length l in
  let sg = Graph.Pack.Digraph.create ~size:n () in
  let sv_of_dp = Hashtbl.create n
  and sdp_of_v = Hashtbl.create n in

  (* Création des sommets *)

  List.iter (fun v ->
               Graph.Pack.Digraph.add_vertex sg v;
               let dep = Hashtbl.find g.dp_of_vertex v in
               Hashtbl.add sv_of_dp dep v;
               Hashtbl.add sdp_of_v v dep) l;

  (* Création des arêtes *)

  List.iter (fun v1 -> List.iter (fun v2 -> if Graph.Pack.Digraph.mem_edge g.graph v1 v2 then Graph.Pack.Digraph.add_edge sg v1 v2) l) l;

  {graph = sg; vertex_of_dp = sv_of_dp; dp_of_vertex = sdp_of_v};;


(* Fonction retirant les graphes n'ayant aucune arête (voir commentaire
   ci-dessous). *)

let remove_no_edge l =

  let rec rec_remove_no_edge l acc =
    match l with
      | [] -> acc
      | g::q when Graph.Pack.Digraph.nb_edges g.graph > 0 -> rec_remove_no_edge q (g::acc)
      | _::q -> rec_remove_no_edge q acc in

  rec_remove_no_edge l [];;


(* Extraction des composantes fortement connexes du graphe *)

(* On utilise la fonction Graph.Pack.Digraph.Components.scc_list, qui ne
   donne pas exactement les composantes connexes au sens du polycopié :
   en effet, elle considère comme une composante fortement connexe un
   sommet isolé avec aucune arête. On élimine donc ces graphes. *)

let extract_components (g:graph) : graph list =
  let l = Graph.Pack.Digraph.Components.scc_list g.graph in
  let l' = List.rev_map (subgraph_of_vertex_list g) l in
  remove_no_edge l';;


(* Réduits d'un terme en un pas par une règle *)

let rec compute_rule_reds r t =
  let (g,d) = r in
  match t with
    | Var _ -> []
    | Term(f,l) ->
        let count = ref (-1) in
        let labeled_l = List.rev_map (fun x -> incr count; (x,!count)) l in
        let sub_res =
          List.flatten (List.rev_map (fun (x,i) -> List.rev_map (fun y -> Term(f,list_replace l i y)) (compute_rule_reds r x)) labeled_l) in
        match unification g t with
          | Some subst -> (substitute subst d)::sub_res
          | None -> sub_res;;


(* Réduits en un pas d'un terme *)

let rec compute_one_step_reds (s:system) (t:term) : term list =
  match s with
    | [] -> []
    | r::q -> List.rev_append (compute_rule_reds r t) (compute_one_step_reds q t);;


(* Réduits en n pas d'un terme *)

let rec compute_n_step_reds (s:system) (n:int) (t:term) : term list =
  match n with
    | 0 -> [t]
    | p -> List.flatten (List.rev_map (compute_one_step_reds s) (compute_n_step_reds s (p-1) t));;


(* Projection simple *)

(* Étant donné le type de retour de find_projection, il me semble plus
   simple qu'une projection simple n'ait pas pour type (string -> int) mais
   ((string*int) list). *)

let project (p:(string*int) list) (t:term) : term =
  match t with
    | Var v -> Var v
    | Term(f,l) -> List.nth l (List.assoc f p);;


(* Système formé des règles sous-terme d'un système passé en argument *)

let syst_sub_term (s:system) : system =
  let symbs = compute_symbs_arit s in
  List.flatten (
    List.rev_map
      (fun (sy,n) -> list_of_fun (fun i -> (Term(sy,list_of_fun (fun i -> Var i) 0 (n-1)),Var i)) 0 (n-1))
      symbs
  );;


(* Test pour savoir si on peut retirer une paire de dépendance *)

let removable (s:system) (p:(string*int) list) (dep:dp) : can_be_removed =
  let s' = List.rev_append s (syst_sub_term s) in
  let (u,v) = dep in
  let u' = project p u
  and v' = project p v in
  let flag = ref true
  and i = ref 0 in

  while !flag & !i<=max_test do
    let l = compute_n_step_reds s' !i u' in
    List.iter (fun t -> if v' = t then flag := false) l;
    incr i
  done;

  if !flag then
    No
  else
    if !i=1 then
      Large
    else
      Strict;;


(* Calcul de toutes les projections associées à un symbole *)

let all_projections_symb t =
  let (f,n) = t in
  list_of_fun (fun i -> (f,i)) 0 (n-1);;


(* Calcul de toutes les projections associées à une liste de symboles *)

let rec all_projections ds =
  match ds with
    | [] -> [[]]
    | t::q -> list_prod (all_projections_symb t) (all_projections q);;


(* Test pour savoir si une projection donnée permet de retirer la paire
   de dépendance passée en argument *)

let good_projection s g dep p =
  let vert_dep = Hashtbl.find g.vertex_of_dp dep in
  let flag = ref true in
  Graph.Pack.Digraph.iter_vertex (fun v ->
                                    if !flag then
                                      if Graph.Pack.Digraph.V.equal vert_dep v then
                                        match removable s p dep with
                                          | Strict -> ()
                                          | _ -> flag := false
                                      else
                                        match removable s p (Hashtbl.find g.dp_of_vertex v) with
                                          | No -> flag := false
                                          | _ -> ()
                                 ) g.graph;

  !flag;;


(* Fonction cherchant une projection simple pour la paire de dépendance
   passée en argument *)

let rec find_projection_dp s g l dep = (* l = all_projections ds *)
  match l with
    | [] -> raise Not_found
    | p::_ when good_projection s g dep p -> p
    | _::q -> find_projection_dp s g q dep;;


(* Trouver la bonne projection simple *)

(* Tel que décrit dans le commentaire de l'interface .mli, je comprends
   que le graphe passé en argument de find_projection est le graphe de
   dépendance en entier, et que l'on ne renvoie qu'une paire de
   dépendance, ce qui correspond donc à une seule composante fortement
   connexe. Je trouve donc plus judicieux que le graphe passé en
   argument soit une composante fortement connexe du graphe de
   dépendance. *)
(* Il me paraît également plus judicieux de passer en argument non pas
   la liste des symboles définis, mais la liste des symboles définis
   avec leur arité. *)

let find_projection (s:system) (g:graph) (ds:(symb*int) list) : (symb * int) list * dp =
  let l = all_projections ds in
  let flag = ref true
  and res_p = ref []
  and res_dp = ref (Var 0, Var 0) in

  Graph.Pack.Digraph.iter_vertex (fun v ->
                                    if !flag then
                                      let dep = Hashtbl.find g.dp_of_vertex v in
                                      try
                                        let p = find_projection_dp s g l dep in
                                        flag := false;
                                        res_p := p;
                                        res_dp := dep
                                      with Not_found -> ()
                                 ) g.graph;

  if !flag then raise Not_found else (!res_p,!res_dp);;


(* Fonction principale *)

(* Tant qu'il reste des composantes fortement connexes à étudier, on
   cherche à supprimer un sommet de l'une d'entre elle:

   - si cela est possible, soit on n'a plus d'arête dans le graphe et on
   s'arrête, soit on a encore des arêtes et on continue pour les
   composante fortement connexes générées;

   - si cela n'est pas possible, on passe à la composante fortement
     connexe suivante.*)

let main (s:system) : graph list =
  let const_symbs = compute_const_symbs s
  and def_symbs_arit = compute_def_symbs_arit s in
  let dp_list = compute_dps s in
  let g = compute_graph const_symbs dp_list in
  let l = extract_components g in

  let rec reduction g_list acc =
    match g_list with
      | [] -> acc
      | g::q ->
          try
            let (_,dep) = find_projection s g def_symbs_arit in
            Graph.Pack.Digraph.remove_vertex g.graph (Hashtbl.find g.vertex_of_dp dep);
            if Graph.Pack.Digraph.nb_edges g.graph = 0 then
              reduction q acc
            else
              let l = extract_components g in
              reduction (List.rev_append l q) acc
          with Not_found -> reduction q (g::acc) in

  reduction l [];;
